#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

The RAMSES-II system (controllers, zones, devices).
"""

__version__ = "0.22.1"
VERSION = __version__
